# Generated by Django 2.2.15 on 2020-09-04 16:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0023_increase_distribution_release_short'),
    ]

    operations = [
        migrations.AlterField(
            model_name='variant',
            name='repository',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='variants', to='core.Repository'),
        ),
    ]
